"""This module contains a class designed to read and present information from
a JSON configuration file. This configuration file provides overriding
information for attributes provided by granules on a per-collection basis.
This information is primarily intended to augment the CF-Convention
attributes for a dataset, but can also be used to alter non CF-Convention
metadata within a granule.

Information within the configuration file is split into rules that have
an Applicability. This section should define a mission, collection short
name and (optionally) a regular expression compatible string for relevant
variable paths.

The configuration file also specifies variables that are incorrectly
considered as science variables by the VarInfo class, due to them having
references in the coordinates attribute. Further, there are required
variables that have to be included in the output of any variable subset
request for a specific collection.

"""

from __future__ import annotations

from os.path import exists
from typing import Any
import json
import re

from varinfo.exceptions import (
    InvalidConfigFileFormatError,
    MissingConfigurationFileError,
)


class CFConfig:
    """This class should read the main configuration file,
    see e.g. sample_config.json, which defines overriding values for the
    attributes stored in fields such as ancillary_variables, or dimensions.

    Given a mission and collection short name, upon instantiation, the
    object should only retain information relevant to that specific
    collection.

    """

    def __init__(
        self,
        mission: str | None,
        collection_short_name: str | None,
        config_file: str | None = None,
    ):
        """Set supplied class attributes. Then read the designated
        configuration file to obtain mission and short name specific
        attributes.

        """
        self.config_file = config_file
        self.mission = mission
        self.short_name = collection_short_name

        self.metadata_overrides: dict[str, dict[str, Any]] = {}
        self.excluded_science_variables: set[str] = set()
        self.required_variables: set[str] = set()

        if self.mission is not None:
            self._read_config_file()

    def _read_config_file(self):
        """Open the main configuration JSON file and extract only those parts
        of it pertaining to the mission and collection specified upon
        instantiating the class.

        """
        if self.config_file is not None and not exists(self.config_file):
            raise MissingConfigurationFileError(self.config_file)
        elif self.config_file is not None and self.config_file.endswith('.json'):
            with open(self.config_file, 'r', encoding='utf-8') as file_handler:
                config = json.load(file_handler)
        elif self.config_file is not None:
            raise InvalidConfigFileFormatError(self.config_file)
        else:
            config = {}

        self.excluded_science_variables = {
            pattern
            for item in config.get('ExcludedScienceVariables', [])
            if self._is_applicable(
                item['Applicability'].get('Mission'),
                item['Applicability'].get('ShortNamePath'),
            )
            for pattern in item['VariablePattern']
        }

        self.required_variables = {
            pattern
            for item in config.get('RequiredVariables', [])
            if self._is_applicable(
                item['Applicability'].get('Mission'),
                item['Applicability'].get('ShortNamePath'),
            )
            for pattern in item['VariablePattern']
        }

        for override in config.get('MetadataOverrides', []):
            self._process_cf_item(override, self.metadata_overrides)

    def _is_applicable(self, mission: str, short_name: str | None = None) -> bool:
        """Given a mission, and optionally also a collection short name, of an
        applicability within the configuration file, check for a match
        against the mission and short name specified when instantiating the
        class object.

        """
        mission_matches = re.match(mission, self.mission) is not None

        short_name_matches = (
            short_name is None or re.match(short_name, self.short_name) is not None
        )

        return mission_matches and short_name_matches

    def _process_cf_item(
        self,
        cf_item: dict,
        results: dict[str, dict],
        input_mission: str | None = None,
        input_short_name: str | None = None,
    ):
        """Process a single block in the `MetadataOverrides` region of the
        configuration file. First check that the applicability matches the
        mission and short name for the class. Next, check for a variable
        pattern. This is indicative of there being overriding attributes in
        this list item. Assign any information to the results dictionary, with
        a key of that variable pattern.

        """
        mission = cf_item['Applicability'].get('Mission') or input_mission
        short_name = cf_item['Applicability'].get('ShortNamePath') or input_short_name

        if mission is not None and self._is_applicable(mission, short_name):
            # Some outer Applicability items have attributes, but no
            # variable path - the assumption here is that the applicability is
            # to all variables (see ICESat2 dimensions override, SPL4.* and
            # SPL3FTA grid_mapping overrides)
            pattern = cf_item['Applicability'].get('VariablePattern', '.*')
            results[pattern] = self._create_attributes_object(cf_item)

    @staticmethod
    def _create_attributes_object(cf_item: dict) -> dict[str, str]:
        """Construct a dictionary object containing all contained attributes,
        which are specified as list items with Name and Value keys.

        """
        return {
            attribute['Name']: attribute['Value']
            for attribute in cf_item.get('Attributes', {})
        }

    def get_metadata_overrides(self, variable_path: str) -> dict[str, Any]:
        """Return the MetadataOverrides that match a given variable. If there
        are no overrides, then empty dictionaries will be returned instead.

        First iterate through the self.metadata_overrides and find all items
        with a variable pattern that matches the supplied variable (or group)
        path.

        Next sort that dictionary, so that matching patterns are:

        * Primarily sorted from shallowed to deepest, by counting the total
          number of slashes in the string.
        * Within each depth (with the same number of slashes), patterns are
          sorted from shortest to longest string length.

        It is assumed that regular expressions that match deeper elements of
        a file hierarchy are intended to be more specifically applied, and that
        within a given depth, the string length is a proxy for specificity.

        Last, combine the attribute names and values from each matching
        override item. Because of the ordering in the previous step, if there
        are multiple values supplied for the same metadata attribute, the value
        retained will be the one with the longest variable pattern, which is a
        proxy for how specific the override is.

        Note: Depth is approximated by counting the _total_ number of slashes
        in the VariablePattern regular expression, and does not account for
        alternation. As such:

        ```
        "Applicability": {
          "VariablePattern": "/(group_one|group_two)/.*
        }
        ```

        will correctly determine a depth of 2, whereas:

        ```
        "Applicability": {
          "VariablePattern": "(/group_one/.*|/group_two/.*)"
        }
        ```
        will incorrectly determine a depth of 4.

        """
        matching_overrides = {
            pattern: attributes
            for pattern, attributes in self.metadata_overrides.items()
            if re.match(pattern, variable_path) is not None
        }

        # Order MetadataOverrides items by:
        # First: Depth of the variable hierarchy included in the regular
        # expression pattern (number of slashes), shallowest to deepest.
        # Second: The length of the variable pattern, from shorted to longest.
        # The second ordering is applied in to regular expressions specifying
        # the same depth of variable hierarchy.
        sorted_overrides = dict(
            sorted(
                matching_overrides.items(),
                key=lambda pattern: (pattern[0].count('/'), len(pattern[0])),
            ),
        )

        # Combine all overrides. In the case a metadata attribute appears in
        # multiple matching overrides, the value that comes later based on the
        # previous sorting of variable patterns will take precedence.
        return {
            attribute_name: attribute_value
            for sorted_override in sorted_overrides.values()
            for attribute_name, attribute_value in sorted_override.items()
        }



"""This module contains classes designed to capture information regarding
metadata attributes as read from either an OPeNDAP DMR or a NetCDF-4 file.
These classes are inherited by both representations of groups and of
variables.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Union
import xml.etree.ElementTree as ET

from netCDF4 import Group as NetCDF4Group
from netCDF4 import Variable as NetCDF4Variable

from varinfo.cf_config import CFConfig
from varinfo.utilities import get_xml_attribute


InputContainerType = Union[ET.Element, NetCDF4Group, NetCDF4Variable]


class AttributeContainerBase(ABC):
    """A class to represent objects that have metadata attributes, such as
    groups or variables within a NetCDF-4 file or OPeNDAP DMR.

    """

    def __init__(
        self,
        container: InputContainerType,
        cf_config: CFConfig,
        namespace: str,
        full_name_path: str,
    ):
        """Extract metadata attributes, including any overrides defined in the
        supplied `CFConfig` instance.

        """
        self.namespace = namespace
        self.full_name_path = full_name_path
        self.metadata_overrides = cf_config.get_metadata_overrides(self.full_name_path)
        self.attributes = self._get_attributes(container)
        self._add_additional_attributes()

    @abstractmethod
    def _get_attributes(self, container: InputContainerType) -> dict[str, Any]:
        """Extract all attributes for the container. The contents of the output
        dictionary will be as stored in the granule metadata, with any
        applicable overrides from `CFConfig`.

        """

    @abstractmethod
    def _get_attribute(self, container: InputContainerType, attribute_name: str) -> Any:
        """Extract an attribute value from the source granule metadata. Any
        applicable overrides from `CFConfig` will be applied before returning
        the attribute value.

        """

    def get_attribute_value(
        self, attribute_name: str, default_value: Any | None = None
    ) -> Any:
        """A convenience function for the end-user to retrieve the value of a
        specified attribute, or use an optional default value if that
        attribute is not present in the container metadata. If no default
        value is supplied, requesting the value of an absent attribute will
        return `None`.

        """
        return self.attributes.get(attribute_name, default_value)

    def _add_additional_attributes(self) -> None:
        """Check the `CFConfig` instance for any metadata attributes that are
        listed, but not included in the original granule metadata. These should
        be added to the variable metadata attributes.

        """
        self._add_missing_attributes(self.metadata_overrides)

    def _add_missing_attributes(self, extra_attributes: dict) -> None:
        """Iterate through a dictionary of attributes from the `CFConfig`
        instance matching this container. If there are any attributes listed
        that are not already present in the `self.attributes` dictionary,
        then add them to with the value from the configuration file.

        """
        for attribute_name, attribute_value in extra_attributes.items():
            if attribute_name not in self.attributes:
                self.attributes[attribute_name] = attribute_value

    def _get_configured_attribute(
        self, attribute_name: str, raw_attribute_value: Any
    ) -> Any:
        """Check the `CFConfig` instance associated with the container for any
        metadata attribute overrides that should be applied to the attribute
        value.

        """
        return self.metadata_overrides.get(attribute_name, raw_attribute_value)


class AttributeContainerFromDmr(AttributeContainerBase):
    """This child class inherits from the `AttributeContainerBase` class and
    implements the abstract methods assuming the container source is part of an
    XML element tree.

    """

    def _get_attributes(self, container: ET.Element) -> dict[str, Any]:
        """Locate all child Attribute elements of the container and extract
        their associated values.

        """
        return {
            attribute.get('name'): self._get_attribute(container, attribute.get('name'))
            for attribute in container.findall(f'{self.namespace}Attribute')
            if attribute.get('name') is not None
        }

    def _get_attribute(self, container: ET.Element, attribute_name: str) -> Any:
        """Extract the value of an XML Attribute element, casting it to the
        appropriate type, applying any necessary metadata overrides.

        """
        raw_value = get_xml_attribute(container, attribute_name, self.namespace)
        return self._get_configured_attribute(attribute_name, raw_value)


class AttributeContainerFromNetCDF4(AttributeContainerBase):
    """This child class inherits from the `AttributeContainerBase` class and
    implements the abstract metadata assuming the container source is part of a
    NetCDF-4 file.

    """

    def _get_attributes(
        self, container: NetCDF4Group | NetCDF4Variable
    ) -> dict[str, Any]:
        """Identify all variable attributes and save them to a dictionary."""
        return {
            attribute_name: self._get_attribute(container, attribute_name)
            for attribute_name in container.ncattrs()
        }

    def _get_attribute(
        self, container: NetCDF4Group | NetCDF4Variable, attribute_name: str
    ) -> Any:
        """Extract the value of the metadata attribute, applying any necessary
        override from the `CFConfig` instance.

        """
        raw_value = container.__dict__.get(attribute_name)
        return self._get_configured_attribute(attribute_name, raw_value)



"""Scaffolding for the earthdata-varinfo package

Note: license and classifier list:
https://pypi.org/pypi?%3Aaction=list_classifiers

"""

from typing import List
import io
import pathlib
import os

from setuptools import find_packages, setup


CURRENT_DIRECTORY = pathlib.Path(__file__).parent.resolve()


def parse_dependencies(file_path: str) -> List[str]:
    """Parse a Pip requirements file, and extract the dependencies."""
    with open(file_path, 'r') as file_handler:
        dependencies = file_handler.read().strip().split('\n')

    return dependencies


def get_readme(current_directory: str) -> str:
    """Parse the README.md in the root of the repository, for the long
    description of this Python package.

    """
    with io.open(
        os.path.join(current_directory, 'README.md'), 'r', encoding='utf-8'
    ) as file_handler:
        readme = file_handler.read()

    return readme


def get_semantic_version(current_directory: str) -> str:
    """Parse the VERSION file in the root of the repository for the semantic
    version number of the version.

    """
    with open(os.path.join(current_directory, 'VERSION'), 'r') as file_handler:
        semantic_version = file_handler.read().strip()

    return semantic_version


setup(
    name='earthdata-varinfo',
    version=get_semantic_version(CURRENT_DIRECTORY),
    author='NASA EOSDIS SDPS Data Services Team',
    author_email='owen.m.littlejohns@nasa.gov',
    description=(
        'A package for parsing Earth Observation science granule '
        'structure and extracting relations between science variables'
        ' and their associated metadata, such as coordinates.'
    ),
    long_description=get_readme(CURRENT_DIRECTORY),
    long_description_content_type='text/markdown',
    url='https://github.com/nasa/earthdata-varinfo',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    include_package_data=True,
    install_requires=parse_dependencies('requirements.txt'),
    extras_require={'dev': parse_dependencies('dev-requirements.txt')},
    test_suite='tests',
    python_requires='>=3.9',
    license='License :: OSI Approved :: Apache Software License',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
)
